/****************************************************************************\
*   VIO - this source file is part of Binary vIEW project (BIEW)             *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1999-2000 Konstantin Boldyshev                     *
*   License:              See below                                          *
*   Author and developer: Konstantin Boldyshev                               *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/generic/linux/vio.c                     *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  general implementation of video i/o functions for linux.                  *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/

#include "__os_dep.h"
#include "_sys_dep.h"
#include "biewlib.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/vt.h>

#define VIO_FILE "/dev/vcsa0"

#define VIO_FAST
#define VIO_MALLOC
#define VMAX_X 200
#define VMAX_Y 100
#define VTMP_LEN 100

int on_console = 0;
unsigned short tvioWidth, tvioHeight, tvioNumColors = 16;
const static int __ansi_color[8] = {0,4,2,6,1,5,3,7};
static int saveX, saveY, firstX, firstY, cursor_type = CUR_NORM, out_fd;

#ifdef VIO_MALLOC
static unsigned char *viomem;
static char *vtmp;
#else
static unsigned char viomem[VMAX_X * VMAX_Y * 2];
static char vtmp[VTMP_LEN];
#endif

static int viohandle;

#define	calc_vioaddr(x,y)	(unsigned char *) (viomem + ((x) + (y) * tvioWidth) * 2)
#define twrite(x)		write(out_fd, (x), strlen(x))

#define _bg(x) ((x) >> 4)
#define _fg(x) ((x) & 0x0f)
#define _2color(x) (char)__ansi_color[((x) & 7)]

static char *_2ansi(unsigned char attr)
{
    sprintf(
    	vtmp,
	"\033[%d;3%d;4%d%sm",
	_fg(attr) > 7  ? 1 : 0,
	_2color(_fg(attr)),
	_2color(_bg(attr)),
	_bg(attr) > 7 ? ";5" : ""
    );
    return vtmp;
}

static void gotoxy(int x,int y)
{
    sprintf(vtmp,"\033[%d;%dH", y + 1, x + 1);
    twrite(vtmp);
}

/*
static unsigned char *calc_vioaddr(int x, int y)
{
    return viomem + (x + y * tvioWidth) * 2;
}
*/

/*

*/

int __vioGetCursorType( void )
{
    return cursor_type;
}

void __vioSetCursorType(int type)
{
    cursor_type = type;

    sprintf(vtmp,"\033[?25%c", type?'h':'l');
    twrite(vtmp);
    if (type) {
	sprintf(vtmp,"\033[?%cc", type == CUR_SOLID ? 7 : 0);
	twrite(vtmp);
    }
}

void __viodGetCursorPos(int *x, int *y)
{
    *x = saveX;
    *y = saveY;
}

void __vioSetCursorPos(int x, int y)
{
    saveX = x;
    saveY = y;
    gotoxy(x, y);
}

void __vioReadBuff(int x, int y, void *buff, unsigned len)
{
    memcpy(buff, calc_vioaddr(x, y), len);
}

void __vioWriteBuff(int x, int y, void *buff, unsigned len)
{
    unsigned char * addr = calc_vioaddr(x, y);

    memcpy(addr, buff, len);

    if (on_console) {
	off_t offset = addr - viomem + 4;

	lseek(viohandle, offset, SEEK_SET);
	write(viohandle, buff, len);
    } else {
	unsigned int i;
	unsigned char c, ca, *p;
	char *pb = malloc(len * 20);

	if (pb == NULL) {
	    printm("Memory allocation failed: %s\nExiting..", strerror(errno));
	    exit(errno);
	}

	*pb = 0;

	gotoxy(x, y);

	for (i = 0; i < len; i += 2) {
	    p = (unsigned char *)buff + i; c = *p; ca = *(p + 1);

	    if (!c) c = ' ';
	    if (c == 0x7f || c == 0x9b ||
		(c < ' ' &&
		(c != 0x07 || c !=0x09 || c !=0x0B || c !=0x18 || c !=0x1A)))
	    c = '.';

/*	    if (c < ' ' c = '.'; */

#ifdef VIO_FAST
	    if ((ca != *(p - 1) && i >= 2) || i == (len - 2) || !i)
#endif
	    strcat(pb, _2ansi(ca));
	    strncat(pb, &c, 1);
	}
	twrite(pb);
	free(pb);
    }
}

void __vioRereadState(void)
{
    if (on_console) {
	lseek(viohandle, 0, SEEK_SET);
	read(viohandle, &tvioHeight, 1);
	read(viohandle, &tvioWidth, 1);
    } else {
	struct winsize w;

	ioctl(out_fd, TIOCGWINSZ, &w);
	tvioWidth = w.ws_col ? w.ws_col : 80;
	tvioHeight = w.ws_row ? w.ws_row : 25;
    }
#ifdef VIO_MALLOC
    viomem = realloc(viomem, tvioWidth * tvioHeight * 2);
#endif
}

void __term_vio(void)
{
    __vioSetCursorType(CUR_NORM);
    __vioSetCursorPos(firstX,firstY);

    if (on_console)
	close(viohandle);
/*    else
	twrite("\033[0m\033[3l");
*/
    close(out_fd);
#ifdef VIO_MALLOC
    free(vtmp);
    free(viomem);
#endif
}

void __init_vio(unsigned long flags)
{
    struct vt_mode vtmode;

    out_fd = open(ttyname(STDOUT_FILENO), O_WRONLY);
    if (out_fd < 0) out_fd = STDOUT_FILENO;

    if ((flags & __TVIO_FLG_DIRECT_CONSOLE_ACCESS) == __TVIO_FLG_DIRECT_CONSOLE_ACCESS)
	on_console = ioctl(out_fd, VT_GETMODE, &vtmode) < 0 ? 0 : 1;

    if (on_console) {
	viohandle = open(VIO_FILE, O_RDWR);
	if (viohandle < 0) {
	    printm("Can't open %s: %s\nDirect console access disabled..\n", VIO_FILE, strerror(errno));
	    on_console = 0;
	}
    }
    if (on_console) {
	read(viohandle, &tvioHeight, 1);
	read(viohandle, &tvioWidth, 1);
	read(viohandle, &firstX, 1);
	read(viohandle, &firstY, 1);
    } else {
	struct winsize w;
	char *term = getenv("TERM");

	if (term == NULL || (
	strcasecmp(term, "linux") &&
	strcasecmp(term, "vt100") &&
	strcasecmp(term, "ansi"))) {
	    printm("Can't handle terminal '%s'\nExiting..\n", term);
	    exit(2);
	}

	ioctl(out_fd, TIOCGWINSZ, &w);
	tvioWidth = w.ws_col ? w.ws_col : 80;
	tvioHeight = w.ws_row ? w.ws_row : 25;
	if (tvioWidth > VMAX_X) tvioWidth = VMAX_X;
	if (tvioHeight > VMAX_Y) tvioHeight = VMAX_Y;
	twrite("\033[0m\033[3h");
    }
    saveX = firstX;
    saveY = firstY;
#ifdef VIO_MALLOC
    vtmp = malloc(VTMP_LEN);
    viomem = malloc(tvioWidth * tvioHeight * 2);
    if (vtmp == NULL || viomem == NULL) {
	printm("Can't allocate memory for output: %s\nExiting..\n", strerror(errno));
	exit(errno);
    }
#endif
    memset(vtmp, 0, VTMP_LEN);
    if (on_console)
    	read(viohandle, viomem, tvioWidth * tvioHeight * 2);
    else
	memset(viomem, 0, tvioWidth * tvioHeight * 2);
}
