/****************************************************************************\
*   KEYBOARD - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1999-2000 Konstantin Boldyshev                     *
*   License:              See below                                          *
*   Author and developer: Konstantin Boldyshev                               *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/generic/unix/keyboard.c                 *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
* ncurses / vt100 keyboard library                                           *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/

#include "__os_dep.h"
#include "kbd_code.h"
#include "biewlib.h"

#include <stdio.h>
#include <string.h>
#include <errno.h>

#ifdef HAVE_NCURSES

#include <curses.h>

#ifdef HAVE_MOUSE
#include <limits.h>
MEVENT me;
static int __mou_nbtns;
#endif

#else

#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <sys/time.h>

static int in_fd;
static struct termios sattr;

#define MAX_SEQ 20	/* max fields number in a category */

typedef struct {
    char c;
    int key;
} _2seq;

typedef struct {
    char pre1;
    char pre2;
    char suf;
    short num;
    _2seq s[MAX_SEQ];
} eseq;

/*
    array of translatable sequences
*/

#define NUM_SEQ 7	/* number of sequence categories */

const static eseq S[NUM_SEQ] = {
{'O', 0, 0, MAX_SEQ, {
    {'A',KE_UPARROW},
    {'B',KE_DOWNARROW},
    {'C',KE_RIGHTARROW},
    {'D',KE_LEFTARROW},
    {'P',KE_F(1)},
    {'Q',KE_F(2)},
    {'R',KE_F(3)},
    {'w',KE_F(3)},
    {'y',KE_F(3)},
    {'S',KE_F(4)},
    {'x',KE_F(4)},
    {'t',KE_F(5)},
    {'v',KE_F(5)},
    {'u',KE_F(6)},
    {'l',KE_F(6)},
    {'q',KE_F(7)},
    {'s',KE_F(7)},
    {'r',KE_F(8)},
    {'p',KE_F(9)},
    {'n',KE_F(9)}}
},
{'[', 0, 0, 8, {
    {'A',KE_UPARROW},
    {'B',KE_DOWNARROW},
    {'C',KE_RIGHTARROW},
    {'D',KE_LEFTARROW},
    {'H',KE_HOME},
    {'K',KE_END},
    {'L',KE_CTL_PGUP},
    {'M',KE_CTL_PGUP}}
},
{'[', '[', 0, 5, {
    {'A',KE_F(1)},
    {'B',KE_F(2)},
    {'C',KE_F(3)},
    {'D',KE_F(4)},
    {'E',KE_F(5)}}
},
{'[', 0, '~', 8, {
    {'1',KE_HOME},
    {'2',KE_INS},
    {'3',KE_DEL},
    {'4',KE_END},
    {'5',KE_PGUP},
    {'6',KE_PGDN},
    {'7',KE_HOME},
    {'8',KE_END}}
},
{'[', '1', '~', 8, {
    {'1',KE_F(1)},
    {'2',KE_F(2)},
    {'3',KE_F(3)},
    {'4',KE_F(4)},
    {'5',KE_F(5)},
    {'7',KE_F(6)},
    {'8',KE_F(7)},
    {'9',KE_F(8)}}
},
{'[', '2', '~', 8, {
    {'0',KE_F(9)},
    {'1',KE_F(10)},
    {'3',KE_SHIFT_F(1)},
    {'4',KE_SHIFT_F(2)},
    {'5',KE_SHIFT_F(3)},
    {'6',KE_SHIFT_F(4)},
    {'8',KE_SHIFT_F(5)},
    {'9',KE_SHIFT_F(6)}}
},
{'[', '3', '~', 4, {
    {'1',KE_SHIFT_F(7)},
    {'2',KE_SHIFT_F(8)},
    {'3',KE_SHIFT_F(9)},
    {'4',KE_SHIFT_F(10)}}
}
};

#ifdef HAVE_SELECT
static int is_key_ready()
{
    struct timeval t = { 0, 0 };
    fd_set readfds;

    FD_ZERO(&readfds);
    FD_SET(in_fd, &readfds);
    return select(in_fd + 1, &readfds, NULL, NULL, &t);
}
#endif

#endif

/*
    keyboard FIFO
*/

#define KBUFSIZE 64

static struct {
    int pool[KBUFSIZE];
    int current;
} keybuf;

static int shift_status = 0;	/* status of shift keys */
int __unix_mou_x = 0, __unix_mou_y = 0, __unix_mou_btns = 0;

static void pushEvent(int event)
{
    int i;

    if (event && keybuf.current < KBUFSIZE) {
#ifdef HAVE_NCURSES
	if (event >= KEY_MIN && event <= KEY_MAX) event = _2B(event);
#endif
	if (keybuf.current)
	    for (i = keybuf.current; i >= 0; i--)
	     keybuf.pool[i + 1] = keybuf.pool[i];
	keybuf.pool[0] = event;
	keybuf.current++;
    }
}

/*
    ReadNextEvent is non-blocking call
*/

void ReadNextEvent(void)
{
    int key = 0;

#define ret(x)	pushEvent(x); return;
#define set_s(x) shift_status &= (x); shift_status ^= (x); ret(KE_SHIFTKEYS);

#ifdef HAVE_NCURSES

    key = getch(); switch(key) {
        case ERR: return;
	case KE_STATUS_ALT	: set_s(KS_ALT);
	case KE_STATUS_SHIFT	: set_s(KS_SHIFT);
	case KE_STATUS_CONTROL	: set_s(KS_CTRL);
	case KE_STATUS_RESET	: set_s(0);
	case KE_ENTER2		: ret(KE_ENTER);
	case KE_C_O		: ret(KE_CTL_(O));
#ifdef HAVE_MOUSE
        case KEY_MOUSE: getmouse(&me);
	                __unix_mou_x = me.x;
			__unix_mou_y = me.y;
			__unix_mou_btns = me.bstate;
			pushEvent(KE_MOUSE);
			return;
#endif
    }

    ret(key);

#else

/*
    VT100
*/

#define _SEQLEN 7

#define get(x) read(in_fd,&(x),1)

    int i;
    char c[_SEQLEN];

#ifdef HAVE_SELECT
    if (!is_key_ready()) return;
#endif

/*
    translate escape sequence
*/

    memset(c, 0, _SEQLEN);

    get(c[0]); switch(c[0]) {
	case 0			: return; /* no key is ready */
	case KE_ESCAPE		: break;
	case KE_STATUS_ALT	: set_s(KS_ALT);
	case KE_STATUS_SHIFT	: set_s(KS_SHIFT);
	case KE_STATUS_CONTROL	: set_s(KS_CTRL);
	case KE_STATUS_RESET	: set_s(0);
	case KE_ENTER2		: ret(KE_ENTER);
	case KE_DEL_BACKSPACE	: ret(KE_BKSPACE);
	case KE_C_O		: ret(KE_CTL_(O));
	default			: key = c[0];
    }

    if (key) goto place_key;

    for (i = 1; i < _SEQLEN - 1; i++) {
#ifdef HAVE_SELECT
	if(is_key_ready()) get(c[i]);
	else break;
#else
	get(c[i]); if(!c[i]) break;
#endif
    }

    if (i < 3) {
	key = c[0];
	goto place_key;
    }

/*
    translate escape sequence
*/

#define S S[i]
	for (i = 0; i < NUM_SEQ && !key; i++)  {
	    int j, n;

 	    if (c[1] != S.pre1) continue;
	    n = 2;
	    if (S.pre2) {
		if (c[n] != S.pre2) continue;
		n++;
	    }
	    for(j = 0; j < S.num; j++)
		if (c[n] == S.s[j].c) {
		    if (S.suf && c[n + 1] != S.suf) break;
		    key = S.s[j].key;
		}
	}

place_key:
	if (key) ret(key);

#endif	/* HAVE_NCURSES */

}

int __kbdGetShiftsKey( void )
{
    return shift_status;
}

int __kbdTestKey(unsigned long flg)
{
#ifdef HAVE_MOUSE
    if(__MsGetBtns() && flg == KBD_NONSTOP_ON_MOUSE_PRESS) return KE_MOUSE;
#endif
    ReadNextEvent();
    return keybuf.current;
}

int __kbdGetKey(unsigned long flg)
{
    int key = 0, s = 0;

#ifdef HAVE_MOUSE
    if (__MsGetBtns() && flg == KBD_NONSTOP_ON_MOUSE_PRESS) return KE_MOUSE;
#endif

    while (!keybuf.current) { __OsYield(); ReadNextEvent(); }
    key = keybuf.pool[--keybuf.current];

    if (!(key == KE_MOUSE || key == KE_SHIFTKEYS)) {
	if ((shift_status & KS_CTRL) == KS_CTRL)	s |= ADD_CONTROL;
	if ((shift_status & KS_ALT) == KS_ALT)		s |= ADD_ALT;
	if ((shift_status & KS_SHIFT) == KS_SHIFT)	s |= ADD_SHIFT;
	shift_status = 0;
    }

    /* CTRL+O */
    if ((key == 'o' || key == 'O') &&
	(shift_status & KS_CTRL) == KS_CTRL) key = KE_CTL_(O);

    return key | s;
}

void __init_keyboard(void)
{
#ifdef HAVE_NCURSES

    raw();
    noecho();
    intrflush(stdscr,FALSE);
    meta(stdscr,TRUE);
    keypad(stdscr,TRUE);
    nodelay(stdscr,TRUE);
/*    notimeout(stdscr,TRUE); */

#else

    struct termios tattr;

    in_fd = open(ttyname(fileno(stdin)), O_RDONLY);
    if (in_fd < 0) in_fd = fileno(stdin);

    tcgetattr(in_fd, &tattr);
    sattr = tattr;
    tattr.c_lflag &= ~(ICANON | ECHO | ISIG);
    tattr.c_cc[VMIN] = 1;
    tattr.c_cc[VTIME] = 0;
    tattr.c_iflag &= ~(INPCK | ISTRIP | IXON);
    tattr.c_oflag |= OPOST | ONLCR;
    tcsetattr(in_fd, TCSANOW, &tattr);

#ifndef HAVE_SELECT
    if (fcntl(in_fd, F_SETFL, fcntl(in_fd, F_GETFL) | O_NONBLOCK) < 0) {
	printm("Can't set nonblocking mode: %s\nExiting..\n", strerror(errno));
	exit(errno);
    }
#endif

#endif	/* HAVE_NCURSES */

#ifdef HAVE_MOUSE
    if ((__mou_nbtns = __init_mouse()) != INT_MAX) __MsSetState(1);
#endif
}

void __term_keyboard(void)
{
#ifdef HAVE_MOUSE
   if(__mou_nbtns != INT_MAX) {
     __MsSetState(0);
     __term_mouse();
   }
#endif

#ifndef HAVE_NCURSES
    tcsetattr(in_fd, TCSANOW, &sattr);
    close(in_fd);
#endif
}
